"""Tests for multiplication, etc using banded matrices."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import unittest
import numpy as np
import random
from numpy.random import randn, randint

import bandmat as bm
import bandmat.full as fl
from bandmat.testhelp import assert_allclose, get_array_mem
from bandmat.test_core import gen_BandMat

def rand_bool():
    return randint(0, 2) == 0

class TestTensor(unittest.TestCase):
    def test_dot_mv_plus_equals(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b = randn(size)
            c = randn(size)
            a_full = a_bm.full()
            c_good = c.copy()
            array_mem = get_array_mem(a_bm.data, b, c)

            bm.dot_mv_plus_equals(a_bm, b, c)
            c_good += np.dot(a_full, b)
            assert_allclose(c, c_good)
            assert get_array_mem(a_bm.data, b, c) == array_mem

    def test_dot_mv(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b = randn(size)
            a_full = a_bm.full()

            c = bm.dot_mv(a_bm, b)
            c_good = np.dot(a_full, b)
            assert_allclose(c, c_good)
            assert not np.may_share_memory(c, a_bm.data)
            assert not np.may_share_memory(c, b)

    def test_dot_mm_plus_equals(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            c_bm = gen_BandMat(size)
            diag = None if rand_bool() else randn(size)
            diag_value = np.ones((size,)) if diag is None else diag
            a_full = a_bm.full()
            b_full = b_bm.full()
            c_full = c_bm.full()
            l = c_bm.l
            u = c_bm.u
            array_mem = get_array_mem(a_bm.data, b_bm.data, c_bm.data)
            if diag is not None:
                diag_mem = get_array_mem(diag)

            bm.dot_mm_plus_equals(a_bm, b_bm, c_bm, diag=diag)
            c_full += fl.band_ec(
                l, u,
                np.dot(np.dot(a_full, np.diag(diag_value)), b_full)
            )
            assert_allclose(c_bm.full(), c_full)
            assert get_array_mem(a_bm.data, b_bm.data, c_bm.data) == array_mem
            if diag is not None:
                assert get_array_mem(diag) == diag_mem

    def test_dot_mm(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            diag = None if rand_bool() else randn(size)
            diag_value = np.ones((size,)) if diag is None else diag
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = bm.dot_mm(a_bm, b_bm, diag=diag)
            c_full = np.dot(np.dot(a_full, np.diag(diag_value)), b_full)
            assert c_bm.l == a_bm.l + b_bm.l
            assert c_bm.u == a_bm.u + b_bm.u
            assert c_bm.size == size
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)
            if diag is not None:
                assert not np.may_share_memory(c_bm.data, diag)

    def test_dot_mm_partial(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            diag = None if rand_bool() else randn(size)
            diag_value = np.ones((size,)) if diag is None else diag
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = bm.dot_mm_partial(l, u, a_bm, b_bm, diag=diag)
            c_full = fl.band_ec(
                l, u,
                np.dot(np.dot(a_full, np.diag(diag_value)), b_full)
            )
            assert c_bm.l == l
            assert c_bm.u == u
            assert c_bm.size == size
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)
            if diag is not None:
                assert not np.may_share_memory(c_bm.data, diag)

    def test_dot_mmm_partial(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            c_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            a_full = a_bm.full()
            b_full = b_bm.full()
            c_full = c_bm.full()

            d_bm = bm.dot_mmm_partial(l, u, a_bm, b_bm, c_bm)
            d_full = fl.band_ec(l, u, np.dot(a_full, np.dot(b_full, c_full)))
            assert d_bm.l == l
            assert d_bm.u == u
            assert d_bm.size == size
            assert_allclose(d_bm.full(), d_full)
            assert not np.may_share_memory(d_bm.data, a_bm.data)
            assert not np.may_share_memory(d_bm.data, b_bm.data)
            assert not np.may_share_memory(d_bm.data, c_bm.data)

    def test_band_of_outer_plus_equals(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_vec = randn(size)
            b_vec = randn(size)
            mult = randn()
            mat_bm = gen_BandMat(size)
            mat_full = mat_bm.full()
            l = mat_bm.l
            u = mat_bm.u
            array_mem = get_array_mem(a_vec, b_vec, mat_bm.data)

            bm.band_of_outer_plus_equals(a_vec, b_vec, mat_bm, mult=mult)
            mat_full += fl.band_ec(l, u, np.outer(a_vec, b_vec) * mult)
            assert_allclose(mat_bm.full(), mat_full)
            assert get_array_mem(a_vec, b_vec, mat_bm.data) == array_mem

    def test_trace_dot(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()

            c = bm.trace_dot(a_bm, b_bm)
            c_good = np.trace(np.dot(a_full.T, b_full))
            assert_allclose(c, c_good)

if __name__ == '__main__':
    unittest.main()



"""Tests for core banded matrix definitions and functions."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import unittest
import doctest
import numpy as np
import random
from numpy.random import randn, randint

import bandmat as bm
import bandmat.full as fl
from bandmat.testhelp import assert_allclose, assert_allequal, get_array_mem

def rand_bool():
    return randint(0, 2) == 0

def gen_BandMat(size, l=None, u=None, transposed=None):
    """Generates a random BandMat."""
    if l is None:
        l = random.choice([0, 1, randint(0, 10)])
    if u is None:
        u = random.choice([0, 1, randint(0, 10)])
    data = randn(l + u + 1, size)
    if transposed is None:
        transposed = rand_bool()
    return bm.BandMat(l, u, data, transposed=transposed)

def load_tests(loader, tests, ignore):
    # package-level doctests (N.B. includes other modules, not just core)
    tests.addTests(doctest.DocTestSuite(bm))
    tests.addTests(doctest.DocTestSuite(bm.core))
    return tests

class TestCore(unittest.TestCase):
    def test_BandMat_basic(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)

            assert a_bm.size == size

    def test_BandMat_full(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            l, u = a_bm.l, a_bm.u

            # N.B. these tests are not really testing much of anything (they
            #   are virtually identical to the implementation of BandMat.full),
            #   but this is not that surprising since the lines below are kind
            #   of the definition of the representation used by BandMat in the
            #   two cases (transposed True and transposed False).
            if a_bm.transposed:
                assert_allequal(a_full.T, fl.band_c(u, l, a_bm.data))
            else:
                assert_allequal(a_full, fl.band_c(l, u, a_bm.data))

            assert not np.may_share_memory(a_full, a_bm.data)

    def test_BandMat_T(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)

            assert_allequal(a_bm.T.full(), a_bm.full().T)

            assert a_bm.T.data is a_bm.data
            if size > 0:
                assert np.may_share_memory(a_bm.T.data, a_bm.data)

    def test_BandMat_copy_exact(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            mat_full_orig = mat_bm.full().copy()

            mat_bm_new = mat_bm.copy_exact()
            assert mat_bm_new.l == mat_bm.l
            assert mat_bm_new.u == mat_bm.u
            assert mat_bm_new.transposed == mat_bm.transposed

            # check that copy represents the same matrix
            assert_allequal(mat_bm_new.full(), mat_full_orig)

            # check that copy does not share memory with original
            assert not np.may_share_memory(mat_bm_new.data, mat_bm.data)

            # check that mutating the copy does not change the original
            mat_bm_new.data += 1.0
            assert_allequal(mat_bm.full(), mat_full_orig)

    def test_BandMat_copy(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            mat_full_orig = mat_bm.full().copy()

            mat_bm_new = mat_bm.copy()
            assert mat_bm_new.l == mat_bm.l
            assert mat_bm_new.u == mat_bm.u
            assert not mat_bm_new.transposed

            # check that copy represents the same matrix
            assert_allequal(mat_bm_new.full(), mat_full_orig)

            # check that copy does not share memory with original
            assert not np.may_share_memory(mat_bm_new.data, mat_bm.data)

            # check that mutating the copy does not change the original
            mat_bm_new.data += 1.0
            assert_allequal(mat_bm.full(), mat_full_orig)

    def test_BandMat_equiv(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            l_new = random.choice([None, 0, 1, randint(0, 10)])
            u_new = random.choice([None, 0, 1, randint(0, 10)])
            transposed_new = random.choice([None, True, False])
            zero_extra = rand_bool()

            l_new_value = mat_bm.l if l_new is None else l_new
            u_new_value = mat_bm.u if u_new is None else u_new
            transposed_new_value = (mat_bm.transposed if transposed_new is None
                                    else transposed_new)

            if l_new_value < mat_bm.l or u_new_value < mat_bm.u:
                self.assertRaises(AssertionError,
                                  mat_bm.equiv,
                                  l_new=l_new, u_new=u_new,
                                  transposed_new=transposed_new,
                                  zero_extra=zero_extra)
            else:
                mat_bm_new = mat_bm.equiv(l_new=l_new, u_new=u_new,
                                          transposed_new=transposed_new,
                                          zero_extra=zero_extra)
                assert mat_bm_new.l == l_new_value
                assert mat_bm_new.u == u_new_value
                assert mat_bm_new.transposed == transposed_new_value
                assert_allequal(mat_bm_new.full(), mat_bm.full())
                assert not np.may_share_memory(mat_bm_new.data, mat_bm.data)

                if zero_extra:
                    mat_new_data_good = (
                        fl.band_e(u_new_value, l_new_value, mat_bm.full().T)
                    ) if mat_bm_new.transposed else (
                        fl.band_e(l_new_value, u_new_value, mat_bm.full())
                    )
                    assert_allequal(mat_bm_new.data, mat_new_data_good)

    def test_BandMat_plus_equals_band_of(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            target_bm = gen_BandMat(size)
            mat_bm = gen_BandMat(size)
            target_full = target_bm.full()
            mat_full = mat_bm.full()
            array_mem = get_array_mem(target_bm.data, mat_bm.data)

            target_bm.plus_equals_band_of(mat_bm, mult)
            target_full += (
                fl.band_ec(target_bm.l, target_bm.u, mat_full) * mult
            )
            assert_allclose(target_bm.full(), target_full)
            assert get_array_mem(target_bm.data, mat_bm.data) == array_mem

    def test_BandMat_add(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = a_bm + b_bm
            c_full = a_full + b_full
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)

            c_bm = a_bm + 0
            c_full = a_full + 0
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)

            c_bm = 0 + a_bm
            c_full = 0 + a_full
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)

            with self.assertRaises(TypeError):
                a_bm + 1.0
            with self.assertRaises(TypeError):
                1.0 + a_bm

    def test_BandMat_sum(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            num_terms = randint(10)
            a_bms = [ gen_BandMat(size) for _ in range(num_terms) ]
            a_fulls = [ a_bm.full() for a_bm in a_bms ]

            b_bm = sum(a_bms)
            b_full = sum(a_fulls)
            if num_terms > 0:
                assert_allclose(b_bm.full(), b_full)
            for a_bm in a_bms:
                assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_sub(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()

            c_bm = a_bm - b_bm
            c_full = a_full - b_full
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)
            assert not np.may_share_memory(c_bm.data, b_bm.data)

            c_bm = a_bm - 0
            c_full = a_full - 0
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)

            c_bm = 0 - a_bm
            c_full = 0 - a_full
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)

            with self.assertRaises(TypeError):
                a_bm - 1.0
            with self.assertRaises(TypeError):
                1.0 - a_bm

    def test_BandMat_iadd(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()
            array_mem = get_array_mem(a_bm.data, b_bm.data)

            if a_bm.l >= b_bm.l and a_bm.u >= b_bm.u:
                a_bm += b_bm
                a_full += b_full
                assert_allclose(a_bm.full(), a_full)
                assert get_array_mem(a_bm.data, b_bm.data) == array_mem
            else:
                with self.assertRaises(AssertionError):
                    a_bm += b_bm

            a_bm += 0
            a_full += 0
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data, b_bm.data) == array_mem

            with self.assertRaises(TypeError):
                a_bm += 1.0

    def test_BandMat_isub(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            b_bm = gen_BandMat(size)
            a_full = a_bm.full()
            b_full = b_bm.full()
            array_mem = get_array_mem(a_bm.data, b_bm.data)

            if a_bm.l >= b_bm.l and a_bm.u >= b_bm.u:
                a_bm -= b_bm
                a_full -= b_full
                assert_allclose(a_bm.full(), a_full)
                assert get_array_mem(a_bm.data, b_bm.data) == array_mem
            else:
                with self.assertRaises(AssertionError):
                    a_bm -= b_bm

            a_bm -= 0
            a_full -= 0
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data, b_bm.data) == array_mem

            with self.assertRaises(TypeError):
                a_bm -= 1.0

    def test_BandMat_pos(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = +a_bm
            b_full = +a_full
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_neg(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = -a_bm
            b_full = -a_full
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_mul_and_rmul(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm * mult
            b_full = a_full * mult
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            c_bm = mult * a_bm
            c_full = mult * a_full
            assert c_bm.l == a_bm.l
            assert c_bm.u == a_bm.u
            assert_allclose(c_bm.full(), c_full)
            assert not np.may_share_memory(c_bm.data, a_bm.data)

            with self.assertRaises(TypeError):
                a_bm * a_bm

    def test_BandMat_various_divs(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm // mult
            b_full = a_full // mult
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm / mult
            b_full = a_full / mult
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm.__floordiv__(mult)
            b_full = a_full.__floordiv__(mult)
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm.__div__(mult)
            b_full = a_full.__div__(mult)
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            b_bm = a_bm.__truediv__(mult)
            b_full = a_full.__truediv__(mult)
            assert b_bm.l == a_bm.l
            assert b_bm.u == a_bm.u
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

            with self.assertRaises(TypeError):
                a_bm // a_bm
            with self.assertRaises(TypeError):
                a_bm / a_bm
            with self.assertRaises(TypeError):
                1.0 // a_bm
            with self.assertRaises(TypeError):
                1.0 / a_bm

    def test_BandMat_imul(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)

            a_bm *= mult
            a_full *= mult
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            with self.assertRaises(TypeError):
                a_bm *= a_bm

    def test_BandMat_various_idivs(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mult = randn()

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm //= mult
            a_full //= mult
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm /= mult
            a_full /= mult
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm.__ifloordiv__(mult)
            a_full.__ifloordiv__(mult)
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm.__idiv__(mult)
            a_full.__idiv__(mult)
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            a_bm = gen_BandMat(size)
            a_full = a_bm.full()
            array_mem = get_array_mem(a_bm.data)
            a_bm.__itruediv__(mult)
            a_full.__itruediv__(mult)
            assert_allclose(a_bm.full(), a_full)
            assert get_array_mem(a_bm.data) == array_mem

            with self.assertRaises(TypeError):
                a_bm //= a_bm
            with self.assertRaises(TypeError):
                a_bm /= a_bm

    def test_BandMat_reverse_view(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm.reverse_view()
            b_full = a_full[::-1, ::-1]
            assert_allclose(b_bm.full(), b_full)
            assert b_bm.data.base is a_bm.data
            if size > 0:
                assert np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_sub_matrix_view(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            start = randint(size + 1)
            end = randint(size + 1)
            if start > end:
                start, end = end, start
            a_bm = gen_BandMat(size)
            a_full = a_bm.full()

            b_bm = a_bm.sub_matrix_view(start, end)
            b_full = a_full[start:end, start:end]
            assert_allclose(b_bm.full(), b_full)
            assert b_bm.data.base is a_bm.data
            if end > start:
                assert np.may_share_memory(b_bm.data, a_bm.data)

    def test_BandMat_embed_as_sub_matrix(self, its=100):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            start = randint(size + 1)
            end = randint(size + 1)
            if start > end:
                start, end = end, start
            a_bm = gen_BandMat(end - start)
            a_full = a_bm.full()

            b_bm = a_bm.embed_as_sub_matrix(start, size)
            b_full = np.zeros((size, size))
            b_full[start:end, start:end] = a_full
            assert_allclose(b_bm.full(), b_full)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_zeros(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            mat_bm = bm.zeros(l, u, size)
            assert mat_bm.l == l
            assert mat_bm.u == u
            assert_allequal(mat_bm.full(), np.zeros((size, size)))

    def test_from_full(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            mat_full = gen_BandMat(size).full()
            zero_outside_band = np.all(fl.band_ec(l, u, mat_full) == mat_full)

            if zero_outside_band:
                mat_bm = bm.from_full(l, u, mat_full)
                assert mat_bm.l == l
                assert mat_bm.u == u
                assert_allequal(mat_bm.full(), mat_full)
                assert not np.may_share_memory(mat_bm.data, mat_full)
            else:
                self.assertRaises(AssertionError, bm.from_full, l, u, mat_full)

    def test_band_c_bm(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])
            mat_rect = randn(l + u + 1, size)

            mat_bm = bm.band_c_bm(l, u, mat_rect)

            mat_full_good = fl.band_c(l, u, mat_rect)
            assert_allequal(mat_bm.full(), mat_full_good)
            assert not np.may_share_memory(mat_bm.data, mat_rect)

    def test_band_e_bm(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            mat_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            mat_rect = bm.band_e_bm(l, u, mat_bm)

            mat_rect_good = fl.band_e(l, u, mat_bm.full())
            assert_allequal(mat_rect, mat_rect_good)
            assert not np.may_share_memory(mat_rect, mat_bm.data)

    def test_band_ec_bm_view(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            b_bm = bm.band_ec_bm_view(l, u, a_bm)

            b_full_good = fl.band_ec(l, u, a_bm.full())
            assert_allequal(b_bm.full(), b_full_good)
            assert b_bm.data.base is a_bm.data
            if size > 0:
                assert np.may_share_memory(b_bm.data, a_bm.data)

    def test_band_ec_bm(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            a_bm = gen_BandMat(size)
            l = random.choice([0, 1, randint(0, 10)])
            u = random.choice([0, 1, randint(0, 10)])

            b_bm = bm.band_ec_bm(l, u, a_bm)

            b_full_good = fl.band_ec(l, u, a_bm.full())
            assert_allequal(b_bm.full(), b_full_good)
            assert not np.may_share_memory(b_bm.data, a_bm.data)

    def test_band_e_bm_common(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])
            num_bms = randint(5)
            mat_bms = [ gen_BandMat(size) for _ in range(num_bms) ]
            if num_bms > 0:
                l = max([ mat_bm.l for mat_bm in mat_bms ])
                u = max([ mat_bm.u for mat_bm in mat_bms ])

            mat_rects = bm.band_e_bm_common(*mat_bms)
            for mat_bm, mat_rect in zip(mat_bms, mat_rects):
                assert_allclose(mat_rect, bm.band_e_bm(l, u, mat_bm))

    def test_diag(self, its=50):
        for it in range(its):
            size = random.choice([0, 1, randint(0, 10), randint(0, 100)])

            vec = randn(size)
            mat_bm = bm.diag(vec)
            assert isinstance(mat_bm, bm.BandMat)
            assert_allequal(mat_bm.full(), np.diag(vec))
            assert mat_bm.data.base is vec
            if size > 0:
                assert np.may_share_memory(mat_bm.data, vec)

            mat_bm = gen_BandMat(size)
            vec = bm.diag(mat_bm)
            assert_allequal(vec, np.diag(mat_bm.full()))
            assert vec.base is mat_bm.data
            if size > 0:
                assert np.may_share_memory(vec, mat_bm.data)

if __name__ == '__main__':
    unittest.main()



"""Functions to do with overlapping subtensors."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import bandmat as bm

import numpy as np

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

@cython.boundscheck(False)
def sum_overlapping_v(cnp.ndarray[cnp.float64_t, ndim=2] contribs,
                      unsigned long step=1,
                      cnp.ndarray[cnp.float64_t, ndim=1] target=None):
    """Computes the overlapped sum of a sequence of vectors.

    The overlapped sum of a sequence of vectors is defined as follows.
    Suppose the vectors in `contribs` are "laid out" along some larger vector
    such that each element of `contribs` is offset by `step` indices relative
    to the previous element.
    For example `contribs[0]` occupies the left edge of the larger vector,
    `contribs[1]` starts at index `step` and `contribs[2]` starts at index
    `step * 2`.
    The overlapped sum is the sum of these vectors laid out in this way, which
    is a vector.

    If `target` is None then a new vector is returned; otherwise the overlapped
    sum is added to the vector `target`.
    If `contribs` has shape `(num_contribs, width)` then the returned vector
    (or `target` if specified) has size `num_contribs * step + width - step`.
    The value of `width` here must be at least `step`.
    """
    assert contribs.shape[1] >= step

    cdef unsigned long width = contribs.shape[1]
    cdef unsigned long num_contribs = contribs.shape[0]
    cdef unsigned long vec_size = num_contribs * step + width - step
    cdef cnp.ndarray[cnp.float64_t, ndim=1] vec

    if target is None:
        vec = np.zeros((vec_size,), dtype=contribs.dtype)
    else:
        assert target.shape[0] == vec_size
        vec = target

    cdef unsigned long index, frame, k

    for index in range(num_contribs):
        frame = index * step
        for k in range(width):
            vec[frame + k] += contribs[index, k]

    if target is None:
        return vec
    else:
        return

@cython.boundscheck(False)
def sum_overlapping_m(cnp.ndarray[cnp.float64_t, ndim=3] contribs,
                      unsigned long step=1,
                      target_bm=None):
    """Computes the overlapped sum of a sequence of square matrices.

    The overlapped sum of a sequence of matrices is defined as follows.
    Suppose the matrices in `contribs` are "laid out" along the diagonal of
    some larger matrix such that each element of `contribs` is `step` indices
    further down and `step` indices further right than the previous element.
    For example `contribs[0]` occupies the top left corner of the larger matrix
    and `contribs[1]` is `step` indices down and right of this.
    The overlapped sum is the sum of these matrices laid out in this way, which
    is a banded matrix.
    For this function the contributions are square, so the resulting banded
    matrix is also square.

    If `target_bm` is None then a new BandMat is returned; otherwise the
    overlapped sum is added to the BandMat `target_bm`.
    If `contribs` has shape `(num_contribs, width, width)` then the returned
    BandMat (or `target_bm` if specified) has upper and lower bandwidth
    `width - 1` and size `num_contribs * step + width - step`.
    The value of `width` here must be at least `step`.
    """
    assert contribs.shape[1] >= 1 and contribs.shape[1] >= step
    assert contribs.shape[2] == contribs.shape[1]

    cdef unsigned long width = contribs.shape[1]
    cdef unsigned long num_contribs = contribs.shape[0]
    cdef unsigned long depth = contribs.shape[1] - 1
    cdef unsigned long mat_size = num_contribs * step + width - step

    if target_bm is None:
        mat_bm = bm.zeros(depth, depth, mat_size)
    else:
        assert target_bm.l == depth and target_bm.u == depth
        assert target_bm.size == mat_size
        mat_bm = target_bm

    cdef unsigned long transposed = mat_bm.transposed
    cdef cnp.ndarray[cnp.float64_t, ndim=2] mat_data = mat_bm.data

    cdef unsigned long index, frame, k, l

    if transposed:
        for index in range(num_contribs):
            frame = index * step
            for k in range(width):
                for l in range(width):
                    mat_data[depth + l - k, frame + k] += contribs[index, k, l]
    else:
        for index in range(num_contribs):
            frame = index * step
            for k in range(width):
                for l in range(width):
                    mat_data[depth + k - l, frame + l] += contribs[index, k, l]

    if target_bm is None:
        return mat_bm
    else:
        return

@cython.boundscheck(False)
def extract_overlapping_v(cnp.ndarray[cnp.float64_t, ndim=1] vec,
                          unsigned long width,
                          unsigned long step=1,
                          cnp.ndarray[cnp.float64_t, ndim=2] target=None):
    """Extracts overlapping subvectors from a vector.

    The result `subvectors` is a matrix consisting of a sequence of subvectors
    of `vec`.
    Specifically `subvectors[i]` is `vec[(i * step):(i * step + width)]`.

    If `target` is None then a new matrix is returned; otherwise the result is
    written to the matrix `target` (and all elements of `target` are guaranteed
    to be overwritten, so there is no need to zero it ahead of time).
    The length of `vec` should be `num_subs * step + width - step` for some
    `num_subs`, i.e. it should "fit" a whole number of subvectors.
    The returned matrix has shape `(num_subs, width)`.
    The value of `width` here must be at least `step`.
    """
    assert step >= 1
    assert width >= step
    assert vec.shape[0] >= width - step
    assert (vec.shape[0] + step - width) % step == 0

    cdef unsigned long num_subs = (vec.shape[0] + step - width) // step
    cdef cnp.ndarray[cnp.float64_t, ndim=2] subvectors

    if target is None:
        subvectors = np.empty((num_subs, width), dtype=vec.dtype)
    else:
        assert target.shape[0] == num_subs
        assert target.shape[1] == width
        subvectors = target

    cdef unsigned long index, frame, k

    for index in range(num_subs):
        frame = index * step
        for k in range(width):
            subvectors[index, k] = vec[frame + k]

    if target is None:
        return subvectors
    else:
        return

@cython.boundscheck(False)
def extract_overlapping_m(mat_bm,
                          unsigned long step=1,
                          cnp.ndarray[cnp.float64_t, ndim=3] target=None):
    """Extracts overlapping submatrices along the diagonal of a banded matrix.

    The result `submats` is rank-3 tensor consisting of a sequence of
    submatrices from along the diagonal of the matrix represented by `mat_bm`.
    The upper and lower bandwidth of `mat_bm` should be the same.
    Let `width` be `mat_bm.l + 1`.
    If `mat_full` is the matrix represented by `mat_bm` then `submats[i]` is
    `mat_full[(i * step):(i * step + width), (i * step):(i * step + width)]`.

    If `target` is None then a new tensor is returned; otherwise the result is
    written to the tensor `target` (and all elements of `target` are guaranteed
    to be overwritten, so there is no need to zero it ahead of time).
    The size of `mat_bm` should be `num_subs * step + width - step` for some
    `num_subs`, i.e. it should "fit" a whole number of submatrices.
    The returned matrix has shape `(num_subs, width, width)`.
    The value of `width` here must be at least `step`.
    """
    assert mat_bm.l == mat_bm.u

    cdef unsigned long width = mat_bm.l + 1

    assert step >= 1
    assert width >= step
    assert mat_bm.size >= width - step
    assert (mat_bm.size + step - width) % step == 0

    cdef unsigned long num_subs = (mat_bm.size + step - width) // step
    cdef unsigned long depth = mat_bm.l
    cdef unsigned long transposed = mat_bm.transposed
    cdef cnp.ndarray[cnp.float64_t, ndim=2] mat_data = mat_bm.data
    cdef cnp.ndarray[cnp.float64_t, ndim=3] submats

    if target is None:
        submats = np.empty((num_subs, width, width), dtype=mat_data.dtype)
    else:
        assert target.shape[0] == num_subs
        assert target.shape[1] == width
        assert target.shape[2] == width
        submats = target

    cdef unsigned long index, frame, k, l

    if transposed:
        for index in range(num_subs):
            frame = index * step
            for k in range(width):
                for l in range(width):
                    submats[index, k, l] = mat_data[depth + l - k, frame + k]
    else:
        for index in range(num_subs):
            frame = index * step
            for k in range(width):
                for l in range(width):
                    submats[index, k, l] = mat_data[depth + k - l, frame + l]

    if target is None:
        return submats
    else:
        return

def sum_overlapping_v_chunked(contribs_chunks, width, target, step=1):
    """A chunked version of sum_overlapping_v.

    The elements of the iterator `contribs_chunks` should be of the form
    `(start, end, contribs)`.
    For example if `contribs_all` has length 10 then

        sum_overlapping_v_chunked([
            (0, 3, contribs_all[0:3]),
            (3, 10, contribs_all[3:10])
        ])

    produces the same result as `sum_overlapping_v(contribs_all)`.
    This can be used to construct more memory-efficient code in some cases
    (though not in the above example).
    """
    assert step >= 0
    overlap = width - step
    assert overlap >= 0

    for start, end, contribs in contribs_chunks:
        sum_overlapping_v(
            contribs,
            step=step,
            target=target[(start * step):(end * step + overlap)]
        )

def sum_overlapping_m_chunked(contribs_chunks, target_bm, step=1):
    """A chunked version of sum_overlapping_m.

    The elements of the iterator `contribs_chunks` should be of the form
    `(start, end, contribs)`.
    For example if `contribs_all` has length 10 then

        sum_overlapping_m_chunked([
            (0, 3, contribs_all[0:3]),
            (3, 10, contribs_all[3:10])
        ])

    produces the same result as `sum_overlapping_m(contribs_all)`.
    This can be used to construct more memory-efficient code in some cases
    (though not in the above example).
    """
    assert step >= 0
    depth = target_bm.l
    assert target_bm.u == depth
    width = depth + 1
    overlap = width - step
    assert overlap >= 0

    for start, end, contribs in contribs_chunks:
        sum_overlapping_m(
            contribs,
            step=step,
            target_bm=target_bm.sub_matrix_view(
                start * step, end * step + overlap
            )
        )

def extract_overlapping_v_chunked(vec, width, chunk_size, step=1):
    """A chunked version of extract_overlapping_v.

    An iterator over chunks of the output of extract_overlapping_v is returned.
    This can be used to construct more memory-efficient code in some cases.
    """
    assert step >= 1
    overlap = width - step
    assert overlap >= 0
    num_subs = (len(vec) - overlap) // step
    assert num_subs * step + overlap == len(vec)
    assert num_subs >= 0
    assert chunk_size >= 1

    for start in range(0, num_subs, chunk_size):
        end = min(start + chunk_size, num_subs)
        subvectors = extract_overlapping_v(
            vec[(start * step):(end * step + overlap)],
            width,
            step=step
        )
        yield start, end, subvectors

def extract_overlapping_m_chunked(mat_bm, chunk_size, step=1):
    """A chunked version of extract_overlapping_m.

    An iterator over chunks of the output of extract_overlapping_m is returned.
    This can be used to construct more memory-efficient code in some cases.
    """
    assert step >= 1
    depth = mat_bm.l
    assert mat_bm.u == depth
    width = depth + 1
    overlap = width - step
    assert overlap >= 0
    num_subs = (mat_bm.size - overlap) // step
    assert num_subs * step + overlap == mat_bm.size
    assert num_subs >= 0
    assert chunk_size >= 1

    for start in range(0, num_subs, chunk_size):
        end = min(start + chunk_size, num_subs)
        submats = extract_overlapping_m(
            mat_bm.sub_matrix_view(start * step, end * step + overlap),
            step=step
        )
        yield start, end, submats



#!/usr/bin/python
"""An example of using the bandmat package for a non-trivial application.

This script demonstrates how the bandmat package can be used to write a simple
but very efficient implementation of the maximum probability speech parameter
generation algorithm used in statistical speech synthesis.
(This algorithm is also sometimes known as maximum likelihood speech parameter
generation, but this is a misnomer since the quantity being maximized is not a
function of the model parameters and is therefore not a likelihood.)
Given a sequence of mean and variance parameters over time, the maximum
probability speech parameter generation algorithm computes the natural
parameters of a corresponding Gaussian distribution over trajectories, and then
computes the mean trajectory using a banded Cholesky decomposition.
The mean trajectory is also the most likely or maximum probability trajectory.
"""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import numpy as np

import bandmat as bm
import bandmat.linalg as bla

def build_win_mats(windows, frames):
    """Builds a window matrix of a given size for each window in a collection.

    `windows` specifies the collection of windows as a sequence of
    `(l, u, win_coeff)` triples, where `l` and `u` are non-negative integers
    specifying the left and right extents of the window and `win_coeff` is an
    array specifying the window coefficients.
    The returned value is a list of window matrices, one for each of the
    windows specified in `windows`.
    Each window matrix is a `frames` by `frames` Toeplitz matrix with lower
    bandwidth `l` and upper bandwidth `u`.
    The non-zero coefficients in each row of this Toeplitz matrix are given by
    `win_coeff`.
    The returned window matrices are stored as BandMats, i.e. using a banded
    representation.
    """
    win_mats = []
    for l, u, win_coeff in windows:
        assert l >= 0 and u >= 0
        assert len(win_coeff) == l + u + 1
        win_coeffs = np.tile(np.reshape(win_coeff, (l + u + 1, 1)), frames)
        win_mat = bm.band_c_bm(u, l, win_coeffs).T
        win_mats.append(win_mat)

    return win_mats

def build_poe(b_frames, tau_frames, win_mats, sdw=None):
    r"""Computes natural parameters for a Gaussian product-of-experts model.

    The natural parameters (b-value vector and precision matrix) are returned.
    The returned precision matrix is stored as a BandMat.
    Mathematically the b-value vector is given as:

        b = \sum_d \transpose{W_d} \tilde{b}_d

    and the precision matrix is given as:

        P = \sum_d \transpose{W_d} \text{diag}(\tilde{tau}_d) W_d

    where $W_d$ is the window matrix for window $d$ as specified by an element
    of `win_mats`, $\tilde{b}_d$ is the sequence over time of b-value
    parameters for window $d$ as given by a column of `b_frames`, and
    $\tilde{\tau}_d$ is the sequence over time of precision parameters for
    window $d$ as given by a column of `tau_frames`.
    """
    if sdw is None:
        sdw = max([ win_mat.l + win_mat.u for win_mat in win_mats ])
    num_windows = len(win_mats)
    frames = len(b_frames)
    assert np.shape(b_frames) == (frames, num_windows)
    assert np.shape(tau_frames) == (frames, num_windows)
    assert all([ win_mat.l + win_mat.u <= sdw for win_mat in win_mats ])

    b = np.zeros((frames,))
    prec = bm.zeros(sdw, sdw, frames)

    for win_index, win_mat in enumerate(win_mats):
        bm.dot_mv_plus_equals(win_mat.T, b_frames[:, win_index], target=b)
        bm.dot_mm_plus_equals(win_mat.T, win_mat, target_bm=prec,
                              diag=tau_frames[:, win_index])

    return b, prec

def simple_example_with_random_parameters():
    windows = [
        (0, 0, np.array([1.0])),
        (1, 1, np.array([-0.5, 0.0, 0.5])),
        (1, 1, np.array([1.0, -2.0, 1.0])),
    ]
    num_windows = len(windows)

    frames = 10
    mean_frames = np.random.randn(frames, num_windows)
    var_frames = np.abs(np.random.randn(frames, num_windows))

    b_frames = mean_frames / var_frames
    tau_frames = 1.0 / var_frames
    win_mats = build_win_mats(windows, frames)
    b, prec = build_poe(b_frames, tau_frames, win_mats)
    mean_traj = bla.solveh(prec, b)
    print 'INPUT'
    print '-----'
    print 'mean parameters over time:'
    print mean_frames
    print 'variance parameters over time:'
    print var_frames
    print
    print 'OUTPUT'
    print '------'
    print 'mean trajectory (= maximum probability trajectory):'
    print mean_traj

if __name__ == '__main__':
    simple_example_with_random_parameters()



"""Assorted helpful functions."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

cimport numpy as cnp
cimport cython

cnp.import_array()
cnp.import_ufunc()

def fancy_plus_equals(cnp.ndarray[cnp.int_t, ndim=1] target_index_seq,
                      cnp.ndarray[cnp.float64_t, ndim=1] source,
                      cnp.ndarray[cnp.float64_t, ndim=1] target):
    """Implements a += method with fancy indexing.

    Does what you might expect
        target[target_index_seq] += source
    to do.
    """
    cdef unsigned long source_size

    source_size = source.shape[0]
    assert target_index_seq.shape[0] == source_size

    cdef unsigned long source_index
    cdef long target_index

    for source_index in range(source_size):
        target_index = target_index_seq[source_index]
        target[target_index] += source[source_index]

    return

def fancy_plus_equals_2d(cnp.ndarray[cnp.int_t, ndim=1] target_index_seq,
                         cnp.ndarray[cnp.float64_t, ndim=2] source,
                         cnp.ndarray[cnp.float64_t, ndim=2] target):
    """Implements a += method with fancy indexing.

    Does what you might expect
        target[target_index_seq] += source
    to do.
    """
    cdef unsigned long source_size
    cdef unsigned long size1

    source_size = source.shape[0]
    assert target_index_seq.shape[0] == source_size
    size1 = source.shape[1]
    assert target.shape[1] == size1

    cdef unsigned long source_index
    cdef long target_index
    cdef unsigned long index1

    for source_index in range(source_size):
        target_index = target_index_seq[source_index]
        for index1 in range(size1):
            target[target_index, index1] += source[source_index, index1]

    return

def fancy_plus_equals_3d(cnp.ndarray[cnp.int_t, ndim=1] target_index_seq,
                         cnp.ndarray[cnp.float64_t, ndim=3] source,
                         cnp.ndarray[cnp.float64_t, ndim=3] target):
    """Implements a += method with fancy indexing.

    Does what you might expect
        target[target_index_seq] += source
    to do.
    """
    cdef unsigned long source_size
    cdef unsigned long size1
    cdef unsigned long size2

    source_size = source.shape[0]
    assert target_index_seq.shape[0] == source_size
    size1 = source.shape[1]
    assert target.shape[1] == size1
    size2 = source.shape[2]
    assert target.shape[2] == size2

    cdef unsigned long source_index
    cdef long target_index
    cdef unsigned long index1
    cdef unsigned long index2

    for source_index in range(source_size):
        target_index = target_index_seq[source_index]
        for index1 in range(size1):
            for index2 in range(size2):
                target[target_index, index1, index2] += (
                    source[source_index, index1, index2]
                )

    return



#!/usr/bin/python
"""A setuptools-based script for distributing and installing bandmat."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import os
import numpy as np
from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.sdist import sdist as _sdist

cython_locs = [
    ('bandmat', 'full'),
    ('bandmat', 'core'),
    ('bandmat', 'tensor'),
    ('bandmat', 'linalg'),
    ('bandmat', 'misc'),
    ('bandmat', 'overlap'),
]

with open('README.rst') as readme_file:
    long_description = readme_file.read()

requires = [ line.rstrip('\n') for line in open('requirements.txt') ]

# see "A note on setup.py" in README.rst for an explanation of the dev file
dev_mode = os.path.exists('dev')

if dev_mode:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize

    class sdist(_sdist):
        """A cythonizing sdist command.

        This class is a custom sdist command which ensures all cython-generated
        C files are up-to-date before running the conventional sdist command.
        """
        def run(self):
            cythonize([ os.path.join(*loc)+'.pyx' for loc in cython_locs ])
            _sdist.run(self)

    cmdclass = {'build_ext': build_ext, 'sdist': sdist}
    ext_modules = [
        Extension('.'.join(loc), [os.path.join(*loc)+'.pyx'],
                  extra_compile_args=['-Wno-unused-but-set-variable', '-O3'],
                  include_dirs=[np.get_include()])
        for loc in cython_locs
    ]
else:
    cmdclass = {}
    ext_modules = [
        Extension('.'.join(loc), [os.path.join(*loc)+'.c'],
                  extra_compile_args=['-Wno-unused-but-set-variable', '-O3'],
                  include_dirs=[np.get_include()])
        for loc in cython_locs
    ]

setup(
    name='bandmat',
    version='0.6',
    description='A banded matrix library for python.',
    url='http://github.com/MattShannon/bandmat',
    author='Matt Shannon',
    author_email='matt.shannon@cantab.net',
    license='3-clause BSD (see License file)',
    packages=['bandmat'],
    install_requires=requires,
    long_description=long_description,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
)



"""Tests for functions to do with overlapping subtensors."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import unittest
import numpy as np
import random
from numpy.random import randn, randint

import bandmat as bm
import bandmat.overlap as bmo
from bandmat.testhelp import assert_allclose, assert_allequal
from bandmat.test_core import gen_BandMat

cc = bm.band_e_bm_common

def rand_bool():
    return randint(0, 2) == 0

def chunk_randomly(xs):
    size = len(xs)

    num_divs = random.choice([0, randint(size // 2 + 1), randint(size + 3)])
    divs = [0] + sorted(
        [ randint(size + 1) for _ in range(num_divs) ]
    ) + [size]

    for start, end in zip(divs, divs[1:]):
        yield start, end, xs[start:end]

class TestOverlap(unittest.TestCase):
    def test_sum_overlapping_v(self, its=50):
        for it in range(its):
            num_contribs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(2), randint(10)])
            width = step + random.choice([0, 1, randint(10)])
            overlap = width - step
            vec_size = num_contribs * step + overlap

            contribs = randn(num_contribs, width)
            target = randn(vec_size)
            target_orig = target.copy()

            vec = bmo.sum_overlapping_v(contribs, step=step)
            assert vec.shape == (vec_size,)

            # check target-based version adds to target correctly
            bmo.sum_overlapping_v(contribs, step=step, target=target)
            assert_allclose(target, target_orig + vec)

            if num_contribs == 0:
                # check action for no contributions
                assert_allequal(vec, np.zeros((overlap,)))
            elif num_contribs == 1:
                # check action for a single contribution
                assert_allequal(vec, contribs[0])
            else:
                # check action under splitting list of contributions in two
                split_pos = randint(num_contribs + 1)
                vec_again = np.zeros((vec_size,))
                bmo.sum_overlapping_v(
                    contribs[:split_pos],
                    step=step,
                    target=vec_again[0:(split_pos * step + overlap)]
                )
                bmo.sum_overlapping_v(
                    contribs[split_pos:],
                    step=step,
                    target=vec_again[(split_pos * step):vec_size]
                )
                assert_allclose(vec, vec_again)

    def test_sum_overlapping_m(self, its=50):
        for it in range(its):
            num_contribs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(2), randint(10)])
            width = max(step, 1) + random.choice([0, 1, randint(10)])
            depth = width - 1
            assert depth >= 0
            overlap = width - step
            mat_size = num_contribs * step + overlap

            contribs = randn(num_contribs, width, width)
            target_bm = gen_BandMat(mat_size, l=depth, u=depth)
            target_bm_orig = target_bm.copy()

            mat_bm = bmo.sum_overlapping_m(contribs, step=step)
            assert mat_bm.size == mat_size
            assert mat_bm.l == mat_bm.u == depth

            # check target-based version adds to target_bm correctly
            bmo.sum_overlapping_m(contribs, step=step, target_bm=target_bm)
            assert_allclose(*cc(target_bm, target_bm_orig + mat_bm))

            if num_contribs == 0:
                # check action for no contributions
                assert_allequal(mat_bm.full(), np.zeros((overlap, overlap)))
            elif num_contribs == 1:
                # check action for a single contribution
                assert_allequal(mat_bm.full(), contribs[0])
            else:
                # check action under splitting list of contributions in two
                split_pos = randint(num_contribs + 1)
                mat_bm_again = bm.zeros(depth, depth, mat_size)
                bmo.sum_overlapping_m(
                    contribs[:split_pos],
                    step=step,
                    target_bm=mat_bm_again.sub_matrix_view(
                        0, split_pos * step + overlap
                    )
                )
                bmo.sum_overlapping_m(
                    contribs[split_pos:],
                    step=step,
                    target_bm=mat_bm_again.sub_matrix_view(
                        split_pos * step, mat_size
                    )
                )
                assert_allclose(*cc(mat_bm, mat_bm_again))

    def test_extract_overlapping_v(self, its=50):
        for it in range(its):
            num_subs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(1, 10)])
            width = step + random.choice([0, 1, randint(10)])
            overlap = width - step
            vec_size = num_subs * step + overlap

            vec = randn(vec_size)
            target = None if rand_bool() else randn(num_subs, width)

            if target is None:
                subvectors = bmo.extract_overlapping_v(vec, width, step=step)
                assert subvectors.shape == (num_subs, width)
            else:
                bmo.extract_overlapping_v(vec, width, step=step, target=target)
                subvectors = target

            for index in range(num_subs):
                assert_allequal(
                    subvectors[index],
                    vec[(index * step):(index * step + width)]
                )

    def test_extract_overlapping_m(self, its=50):
        for it in range(its):
            num_subs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(1, 10)])
            width = step + random.choice([0, 1, randint(10)])
            depth = width - 1
            assert depth >= 0
            overlap = width - step
            mat_size = num_subs * step + overlap

            mat_bm = gen_BandMat(mat_size, l=depth, u=depth)
            target = None if rand_bool() else randn(num_subs, width, width)

            if target is None:
                submats = bmo.extract_overlapping_m(mat_bm, step=step)
                assert submats.shape == (num_subs, width, width)
            else:
                bmo.extract_overlapping_m(mat_bm, step=step, target=target)
                submats = target

            for index in range(num_subs):
                assert_allequal(
                    submats[index],
                    mat_bm.sub_matrix_view(
                        index * step, index * step + width
                    ).full()
                )

    def test_sum_overlapping_v_chunked(self, its=50):
        for it in range(its):
            num_contribs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(2), randint(10)])
            width = step + random.choice([0, 1, randint(10)])
            overlap = width - step
            vec_size = num_contribs * step + overlap

            contribs = randn(num_contribs, width)
            contribs_chunks = chunk_randomly(contribs)
            target = randn(vec_size)
            target_orig = target.copy()

            bmo.sum_overlapping_v_chunked(
                contribs_chunks, width, target, step=step
            )
            vec_good = bmo.sum_overlapping_v(contribs, step=step)
            assert_allclose(target, target_orig + vec_good)

    def test_sum_overlapping_m_chunked(self, its=50):
        for it in range(its):
            num_contribs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(2), randint(10)])
            width = max(step, 1) + random.choice([0, 1, randint(10)])
            depth = width - 1
            assert depth >= 0
            overlap = width - step
            mat_size = num_contribs * step + overlap

            contribs = randn(num_contribs, width, width)
            contribs_chunks = chunk_randomly(contribs)
            target_bm = gen_BandMat(mat_size, l=depth, u=depth)
            target_bm_orig = target_bm.copy()

            bmo.sum_overlapping_m_chunked(
                contribs_chunks, target_bm, step=step
            )
            mat_bm_good = bmo.sum_overlapping_m(contribs, step=step)
            assert_allclose(*cc(target_bm, target_bm_orig + mat_bm_good))

    def test_extract_overlapping_v_chunked(self, its=50):
        for it in range(its):
            num_subs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(1, 10)])
            width = step + random.choice([0, 1, randint(10)])
            overlap = width - step
            vec_size = num_subs * step + overlap
            chunk_size = random.choice([1, randint(1, 10), randint(1, 10)])

            vec = randn(vec_size)

            indices_remaining = set(range(num_subs))
            subvectors_all = np.empty((num_subs, width))
            for start, end, subvectors in bmo.extract_overlapping_v_chunked(
                vec, width, chunk_size, step=step
            ):
                assert end >= start + 1
                for index in range(start, end):
                    assert index in indices_remaining
                    indices_remaining.remove(index)
                subvectors_all[start:end] = subvectors

            subvectors_good = bmo.extract_overlapping_v(vec, width, step=step)
            assert_allclose(subvectors_all, subvectors_good)

    def test_extract_overlapping_m_chunked(self, its=50):
        for it in range(its):
            num_subs = random.choice([0, 1, randint(10), randint(100)])
            step = random.choice([1, randint(1, 10)])
            width = step + random.choice([0, 1, randint(10)])
            depth = width - 1
            assert depth >= 0
            overlap = width - step
            mat_size = num_subs * step + overlap
            chunk_size = random.choice([1, randint(1, 10), randint(1, 10)])

            mat_bm = gen_BandMat(mat_size, l=depth, u=depth)

            indices_remaining = set(range(num_subs))
            submats_all = np.empty((num_subs, width, width))
            for start, end, submats in bmo.extract_overlapping_m_chunked(
                mat_bm, chunk_size, step=step
            ):
                assert end >= start + 1
                for index in range(start, end):
                    assert index in indices_remaining
                    indices_remaining.remove(index)
                submats_all[start:end] = submats

            submats_good = bmo.extract_overlapping_m(mat_bm, step=step)
            assert_allclose(submats_all, submats_good)

if __name__ == '__main__':
    unittest.main()



"""Tests for assorted helpful functions."""

# Copyright 2013, 2014, 2015, 2016, 2017 Matt Shannon

# This file is part of bandmat.
# See `License` for details of license and warranty.

import unittest
import random
from numpy.random import randn, randint

from bandmat.misc import fancy_plus_equals, fancy_plus_equals_2d
from bandmat.misc import fancy_plus_equals_3d
from bandmat.testhelp import assert_allclose, get_array_mem

def rand_bool():
    return randint(0, 2) == 0

class TestMisc(unittest.TestCase):
    def test_fancy_plus_equals(self, its=100):
        for it in range(its):
            source_size = random.choice([0, 1, randint(10), randint(100)])
            target_size = random.choice([1, randint(1, 10), randint(1, 100)])
            source = randn(source_size)
            target = randn(target_size)
            target_index_seq = randint(target_size, size=source_size)
            array_mem = get_array_mem(target_index_seq, source, target)

            target_good = target.copy()
            for source_index, target_index in enumerate(target_index_seq):
                target_good[target_index] += source[source_index]

            fancy_plus_equals(target_index_seq, source, target)

            assert_allclose(target, target_good)
            assert get_array_mem(target_index_seq, source, target) == array_mem

    def test_fancy_plus_equals_fails(self, its=100):
        for it in range(its):
            source_size = random.choice([1, randint(1, 10), randint(1, 100)])
            target_size = random.choice([1, randint(1, 10), randint(1, 100)])
            source = randn(source_size)
            target = randn(target_size)
            target_index_seq = randint(target_size, size=source_size)
            target_index_seq[randint(source_size)] = (
                (-target_size - 1 - randint(10)) if rand_bool()
                else target_size + randint(10)
            )
            self.assertRaises(IndexError,
                              fancy_plus_equals,
                              target_index_seq, source, target)

    def test_fancy_plus_equals_2d(self, its=100):
        for it in range(its):
            source_size = random.choice([0, 1, randint(10), randint(100)])
            target_size = random.choice([1, randint(1, 10), randint(1, 100)])
            size1 = random.choice([0, 1, randint(10)])
            source = randn(source_size, size1)
            target = randn(target_size, size1)
            target_index_seq = randint(target_size, size=source_size)
            array_mem = get_array_mem(target_index_seq, source, target)

            target_good = target.copy()
            for source_index, target_index in enumerate(target_index_seq):
                target_good[target_index] += source[source_index]

            fancy_plus_equals_2d(target_index_seq, source, target)

            assert_allclose(target, target_good)
            assert get_array_mem(target_index_seq, source, target) == array_mem

    def test_fancy_plus_equals_2d_fails(self, its=100):
        for it in range(its):
            source_size = random.choice([1, randint(1, 10), randint(1, 100)])
            target_size = random.choice([1, randint(1, 10), randint(1, 100)])
            size1 = random.choice([1, randint(1, 10)])
            source = randn(source_size, size1)
            target = randn(target_size, size1)
            target_index_seq = randint(target_size, size=source_size)
            target_index_seq[randint(source_size)] = (
                (-target_size - 1 - randint(10)) if rand_bool()
                else target_size + randint(10)
            )
            self.assertRaises(IndexError,
                              fancy_plus_equals_2d,
                              target_index_seq, source, target)

    def test_fancy_plus_equals_3d(self, its=100):
        for it in range(its):
            source_size = random.choice([0, 1, randint(10), randint(100)])
            target_size = random.choice([1, randint(1, 10), randint(1, 100)])
            size1 = random.choice([0, 1, randint(10)])
            size2 = random.choice([0, 1, randint(10)])
            source = randn(source_size, size1, size2)
            target = randn(target_size, size1, size2)
            target_index_seq = randint(target_size, size=source_size)
            array_mem = get_array_mem(target_index_seq, source, target)

            target_good = target.copy()
            for source_index, target_index in enumerate(target_index_seq):
                target_good[target_index] += source[source_index]

            fancy_plus_equals_3d(target_index_seq, source, target)

            assert_allclose(target, target_good)
            assert get_array_mem(target_index_seq, source, target) == array_mem

    def test_fancy_plus_equals_3d_fails(self, its=100):
        for it in range(its):
            source_size = random.choice([1, randint(1, 10), randint(1, 100)])
            target_size = random.choice([1, randint(1, 10), randint(1, 100)])
            size1 = random.choice([1, randint(1, 10)])
            size2 = random.choice([1, randint(1, 10)])
            source = randn(source_size, size1, size2)
            target = randn(target_size, size1, size2)
            target_index_seq = randint(target_size, size=source_size)
            target_index_seq[randint(source_size)] = (
                (-target_size - 1 - randint(10)) if rand_bool()
                else target_size + randint(10)
            )
            self.assertRaises(IndexError,
                              fancy_plus_equals_3d,
                              target_index_seq, source, target)

if __name__ == '__main__':
    unittest.main()
